
--[[

   Written by Francois Fleuret (francois@fleuret.org)

   This is free and unencumbered software released into the public
   domain.

   Anyone is free to copy, modify, publish, use, compile, sell, or
   distribute this software, either in source code form or as a
   compiled binary, for any purpose, commercial or non-commercial, and
   by any means.

   In jurisdictions that recognize copyright laws, the author or
   authors of this software dedicate any and all copyright interest in
   the software to the public domain. We make this dedication for the
   benefit of the public at large and to the detriment of our heirs
   and successors. We intend this dedication to be an overt act of
   relinquishment in perpetuity of all present and future rights to
   this software under copyright law.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
   NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   For more information, please refer to <http://unlicense.org/>

]]--

require 'torch'
require 'nn'
require 'sys'

profiler = {}

profiler.color = true

profiler.colors = function(name)
   if profiler.color then
      return sys.COLORS[name]
   else
      return ''
   end
end

function profiler.decorate(model, functionsToDecorate)

   local functionsToDecorate = functionsToDecorate or
      {
         'updateOutput',
         'backward'
      }

   for _, name in pairs(functionsToDecorate) do
      model.accTime = {}

      local nameOrig = name .. '__orig'

      -- We decorate the class and not the object, otherwise we cannot
      -- save models anymore.

      if rawget(model, name) then
         error('We decorate the classes, not the objects, and there is a `'
                  .. name
                  .. '\' function in '
                  .. tostring(model))
      end

      local toDecorate = model

      while not rawget(toDecorate, name) do
         toDecorate = getmetatable(toDecorate)
      end

      if not toDecorate[nameOrig] then
         -- print('Decorating ' .. toDecorate.__typename .. '.' .. name)
         toDecorate[nameOrig] = toDecorate[name]
         toDecorate[name] = function(self, ...)
            local startTime = sys.clock()
            local result = { self[nameOrig](self, unpack({...})) }
            local endTime = sys.clock()
            self.accTime[name] = (self.accTime[name] or 0) + endTime - startTime
            return unpack(result)
         end
      end

   end

   if torch.isTypeOf(model, nn.Container) then
      for _, m in ipairs(model.modules) do
         profiler.decorate(m, functionsToDecorate)
      end
   end

end

function profiler.timing(l, t, nbSamples, totalTime)
   local s

   s = string.format('%s %.02fs %s[%.02f%%]',
                     l, t,
                     profiler.colors('blue'),
                     100 * t / totalTime
   )

   if nbSamples then
      s = s .. string.format(profiler.colors('green') .. ' (%.01fmus/sample)', 1e6 * t / nbSamples)
   end

   s = s .. profiler.colors('black')

   return s
end

function profiler.print(model, nbSamples, totalTime, indent)
   local indent = indent or ''
   local hint

   if not model.accTime then
      error('The model does not seem decorated for profiling.')
   end

   local localTotal = 0
   for _, t in pairs(model.accTime) do
      localTotal = localTotal + t
   end

   totalTime = totalTime or localTotal

   if torch.isTypeOf(model, nn.Container) then
      hint = ' '
   else
      if profiler.color then
         hint = ' '
      else
         hint = '*'
      end
      hint = hint .. profiler.colors('red')
   end

   print(profiler.timing(indent .. hint .. ' ' .. model.__typename,
                         localTotal, nbSamples, totalTime))

   for l, t in pairs(model.accTime) do
      print(profiler.timing(indent .. '  :' .. l, t, nbSamples, totalTime))
   end

   print()

   if torch.isTypeOf(model, nn.Container) then
      for _, m in ipairs(model.modules) do
         profiler.print(m, nbSamples, totalTime, indent .. '  ')
      end
   end
end

return profiler
